"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.increaseInboxMessageFinishedAttempts = exports.markInboxMessageCompleted = exports.initiateInboxMessageProcessing = exports.startedAttemptsIncrement = exports.initializeInboxMessageStorage = void 0;
const pg_1 = require("pg");
const error_1 = require("../common/error");
const utils_1 = require("../common/utils");
/**
 * Initialize the inbox message storage to store incoming messages in the inbox table.
 * @param config The configuration object that defines the values on how to connect to the database and general settings.
 * @param logger A logger instance for logging trace up to error logs
 * @returns The function to store the inbox message data to the database and the shutdown action.
 */
const initializeInboxMessageStorage = (config, logger) => {
    const pool = new pg_1.Pool(config.pgConfig);
    pool.on('error', (err) => {
        logger.error((0, error_1.ensureExtendedError)(err, 'DB_ERROR'), 'PostgreSQL pool error');
    });
    /**
     * The function to store the inbox message data to the database.
     * @param message The received message that should be stored as inbox message
     * @throws Error if the inbox message could not be stored
     */
    return [
        (message) => __awaiter(void 0, void 0, void 0, function* () {
            try {
                yield (0, utils_1.executeTransaction)(yield pool.connect(), (client) => __awaiter(void 0, void 0, void 0, function* () {
                    yield insertInbox(message, client, config, logger);
                }), utils_1.IsolationLevel.ReadCommitted);
            }
            catch (err) {
                const messageError = new error_1.MessageError(`Could not store the inbox message with id ${message.id}`, 'INBOX_ERROR_STORAGE_FAILED', message, err);
                logger.error(messageError, 'Could not store the inbox message');
                throw messageError;
            }
        }),
        () => __awaiter(void 0, void 0, void 0, function* () {
            pool.removeAllListeners();
            try {
                yield pool.end();
            }
            catch (e) {
                logger.warn((0, error_1.ensureExtendedError)(e, 'DB_ERROR'), 'Inbox message storage shutdown error');
            }
        }),
    ];
};
exports.initializeInboxMessageStorage = initializeInboxMessageStorage;
/**
 * This function increases the "started_attempts" for the inbox message by one
 * in the inbox table. This number can then be compared to the
 * "finished_attempts" number which is only increased when a message processing
 * exception was correctly handled or an error was caught. A difference between
 * the two can only happen if the service crashes after increasing the
 * "started_attempts" but before successfully marking the message as done
 * (success case) or catching an error (error case). If the "started_attempts"
 * and the "finished_attempts" field differ by more than one, the chances are
 * high that this message is causing a service crash.
 * It sets the started_attempts, finished_attempts and processed_at
 * values on the inbox message.
 * For additional safety it makes sure, that the inbox item was not and is not
 * currently being worked on.
 * @param message The inbox message for which to acquire a lock and increment the started_attempts
 * @param client The database client. Must be part of a transaction that runs before the message handling transaction.
 * @param config The configuration settings that defines inbox database schema.
 * @returns 'INBOX_MESSAGE_NOT_FOUND' if the message was not found, 'ALREADY_PROCESSED' if it was processed, and otherwise assigns the properties to the message and returns true.
 */
const startedAttemptsIncrement = (message, client, { settings }) => __awaiter(void 0, void 0, void 0, function* () {
    // Use a NOWAIT select to fully lock and immediately fail if another process is locking that inbox row
    const updateResult = yield client.query(
    /* sql*/ `
      UPDATE ${settings.dbSchema}.${settings.dbTable} SET started_attempts = started_attempts + 1 WHERE id IN
        (SELECT id FROM ${settings.dbSchema}.${settings.dbTable} WHERE id = $1 FOR UPDATE NOWAIT)
        RETURNING started_attempts, finished_attempts, processed_at;`, [message.id]);
    if (updateResult.rowCount === 0) {
        return 'INBOX_MESSAGE_NOT_FOUND';
    }
    const { started_attempts, finished_attempts, processed_at } = updateResult.rows[0];
    if (processed_at) {
        return 'ALREADY_PROCESSED';
    }
    // set the values for the poisonous message strategy
    message.startedAttempts = started_attempts;
    message.finishedAttempts = finished_attempts;
    message.processedAt = processed_at;
    return true;
});
exports.startedAttemptsIncrement = startedAttemptsIncrement;
/**
 * This function makes sure the inbox item was not and is not currently being
 * worked on and acquires a lock to prevent other processes to work with this
 * inbox message. As the inbox listener does not run in the same transaction as
 * the message handler code there is a small chance that the handler code
 * succeeds but the WAL inbox message was not acknowledged. This takes care of
 * such cases by ensuring that the message was not yet processed.
 * It sets the started_attempts, finished_attempts and processed_at
 * values (again) on the inbox message to be sure no other process altered them.
 * @param message The inbox message for which to acquire a lock and fill the started_attempts, finished_attempts and processed_at values
 * @param client The database client. Must be part of the transaction where the message handling changes are later done.
 * @param config The configuration settings that defines inbox database schema.
 * @returns 'INBOX_MESSAGE_NOT_FOUND' if the message was not found, 'ALREADY_PROCESSED' if it was processed, and otherwise assigns the properties to the message and returns true.
 */
const initiateInboxMessageProcessing = (message, client, { settings }) => __awaiter(void 0, void 0, void 0, function* () {
    // Use a NOWAIT select to immediately fail if another process is locking that inbox row
    const selectResult = yield client.query(
    /* sql*/ `
    SELECT started_attempts, finished_attempts, processed_at FROM ${settings.dbSchema}.${settings.dbTable} WHERE id = $1 FOR UPDATE NOWAIT;`, [message.id]);
    if (selectResult.rowCount === 0) {
        return 'INBOX_MESSAGE_NOT_FOUND';
    }
    const { started_attempts, finished_attempts, processed_at } = selectResult.rows[0];
    // ensures latest values (e.g. if the `startedAttemptsIncrement` was not called or another
    // process changed them between the `startedAttemptsIncrement` and this call.
    message.startedAttempts = started_attempts;
    message.finishedAttempts = finished_attempts;
    message.processedAt = processed_at;
    if (processed_at) {
        return 'ALREADY_PROCESSED';
    }
    return true;
});
exports.initiateInboxMessageProcessing = initiateInboxMessageProcessing;
/**
 * Marks the inbox message as handled by setting the processed_at date to the current date and time.
 * It also increases the finished_attempts count by one.
 * @param message The inbox message to acknowledge.
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema.
 */
const markInboxMessageCompleted = ({ id }, client, { settings }) => __awaiter(void 0, void 0, void 0, function* () {
    yield client.query(
    /* sql*/ `UPDATE ${settings.dbSchema}.${settings.dbTable} SET processed_at = $1, finished_attempts = finished_attempts + 1 WHERE id = $2`, [new Date().toISOString(), id]);
});
exports.markInboxMessageCompleted = markInboxMessageCompleted;
/**
 * Does increase the finished_attempts count by one.
 * @param message The inbox message to NOT acknowledge.
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema.
 */
const increaseInboxMessageFinishedAttempts = ({ id }, client, { settings }) => __awaiter(void 0, void 0, void 0, function* () {
    yield client.query(
    /* sql*/ `
    UPDATE ${settings.dbSchema}.${settings.dbTable} SET finished_attempts = finished_attempts + 1 WHERE id = $1;`, [id]);
});
exports.increaseInboxMessageFinishedAttempts = increaseInboxMessageFinishedAttempts;
const insertInbox = (message, dbClient, { settings }, logger) => __awaiter(void 0, void 0, void 0, function* () {
    const { id, aggregateType, aggregateId, messageType, payload, metadata, createdAt, } = message;
    const inboxResult = yield dbClient.query(
    /* sql*/ `
    INSERT INTO ${settings.dbSchema}.${settings.dbTable}
      (id, aggregate_type, aggregate_id, message_type, payload, metadata, created_at)
      VALUES ($1, $2, $3, $4, $5, $6, $7)
      ON CONFLICT (id) DO NOTHING`, [id, aggregateType, aggregateId, messageType, payload, metadata, createdAt]);
    if (!inboxResult.rowCount || inboxResult.rowCount < 1) {
        logger.warn(message, `The message with id ${id} already existed`);
    }
});
//# sourceMappingURL=inbox-message-storage.js.map